@@ -1,10 +1,18 @@
 package Prometheus::Tiny;
-$Prometheus::Tiny::VERSION = '0.002';
+$Prometheus::Tiny::VERSION = '0.003';
 # ABSTRACT: A tiny Prometheus client
 
 use warnings;
 use strict;
 
+my $DEFAULT_BUCKETS = [
+               0.005,
+  0.01, 0.025, 0.05, 0.075,
+  0.1,  0.25,  0.5,  0.75,
+  1.0,  2.5,   5.0,  7.5,
+  10
+];
+
 sub new {
   my ($class) = @_;
   return bless {
@@ -19,8 +27,11 @@ sub _format_labels {
 }
 
 sub set {
-  my ($self, $name, $value, $labels) = @_;
-  $self->{metrics}{$name}{$self->_format_labels($labels)} = $value;
+  my ($self, $name, $value, $labels, $timestamp) = @_;
+  my $f_label = $self->_format_labels($labels);
+  $self->{metrics}{$name}{$f_label} = $value;
+  $self->{meta}{timestamp}{$name}{$f_label} = $timestamp
+  	if $timestamp;
   return;
 }
 
@@ -40,6 +51,23 @@ sub dec {
   return $self->add($name, -1, $labels);
 }
 
+sub histogram_observe {
+  my ($self, $name, $value, $labels) = @_;
+
+  $self->inc($name.'_count', $labels);
+  $self->add($name.'_sum', $value, $labels);
+
+  my @buckets = @{$self->{meta}{$name}{buckets} || $DEFAULT_BUCKETS};
+
+  my $bucket_metric = $name.'_bucket';
+  for my $bucket (@buckets) {
+    $self->add($bucket_metric, $value <= $bucket ? 1 : 0, { %{$labels || {}} , le => $bucket });
+  }
+  $self->inc($bucket_metric, { %{$labels || {}}, le => '+Inf' });
+
+  return;
+}
+
 sub declare {
   my ($self, $name, %meta) = @_;
   $self->{meta}{$name} = { %meta };
@@ -57,10 +85,25 @@ sub format {
       (defined $self->{meta}{$name}{type} ?
         ("# TYPE $name $self->{meta}{$name}{type}\n") : ()),
       (map {
+	  	my $ts = $self->{meta}{timestamp}{$name}{$_} ? ' '.$self->{meta}{timestamp}{$name}{$_} : '';
         $_ ?
-          join '', $name, '{', $_, '} ', $self->{metrics}{$name}{$_}, "\n" :
-          join '', $name, ' ', $self->{metrics}{$name}{$_}, "\n"
-      } sort keys %{$self->{metrics}{$name}}),
+          join '', $name, '{', $_, '} ', $self->{metrics}{$name}{$_}, $ts, "\n" :
+          join '', $name, ' ', $self->{metrics}{$name}{$_}, $ts, "\n"
+      } sort {
+        $name =~ m/_bucket$/ ?
+          do {
+            my $t_a = $a; $t_a =~ s/le="([^"]+)"//; my $le_a = $1;
+            my $t_b = $b; $t_b =~ s/le="([^"]+)"//; my $le_b = $1;
+            $t_a eq $t_b ?
+              do {
+                $le_a eq '+Inf' ? 1 :
+                $le_b eq '+Inf' ? -1 :
+                ($a cmp $b)
+              } :
+              ($a cmp $b)
+          } :
+          ($a cmp $b)
+      } keys %{$self->{metrics}{$name}}),
     )
   } sort keys %names;
 }
@@ -137,9 +180,9 @@ L<Prometheus::Tiny::Shared> for that!
 
 =head2 set
 
-    $prom->set($name, $value, { labels })
+    $prom->set($name, $value, { labels }, [timestamp])
 
-Set the value for the named metric. The labels hashref is optional.
+Set the value for the named metric. The labels hashref is optional. The timestamp (milliseconds since epoch) is optional, but requires labels to be provided to use. An empty hashref will work in the case of no labels.
 
 =head2 add
 
@@ -163,12 +206,23 @@ A shortcut for
 
     $prom->add($name, -1, { labels })
 
+=head2 histogram_observe
+
+    $prom->histogram_observe($name, $value, { labels })
+
+Record a histogram observation. The labels hashref is optional.
+
 =head2 declare
 
-    $prom->declare($name, help => $help, type => $type)
+    $prom->declare($name, help => $help, type => $type, buckets => [...])
 
 "Declare" a metric by setting its help text or type.
 
+For histogram metrics, you can optionally specify the buckets to use. If you
+don't, and later call C<histogram_observe>, the following buckets will be used:
+
+    [ 0.005, 0.01, 0.025, 0.05, 0.075, 0.1, 0.25, 0.5, 0.75, 1.0, 2.5, 5.0, 7.5, 10 ]
+
 =head2 format
 
     my $metrics = $prom->format
@@ -222,6 +276,16 @@ Rob N ★ <robn@robn.io>
 
 =back
 
+=head1 CONTRIBUTORS
+
+=over 4
+
+=item *
+
+ben hengst <ben.hengst@dreamhost.com>
+
+=back
+
 =head1 COPYRIGHT AND LICENSE
 
 This software is copyright (c) 2017 by Rob N ★
